import json
import os
from pathlib import Path
from typing import Dict

import charset_normalizer
import pandas as pd
from cdislogging import get_logger

from heal.vlmd.config import CSV_SCHEMA, JSON_SCHEMA

logger = get_logger("validate-utils", log_level="info")


def detect_file_encoding(file_path):
    """
    detects file encoding using charset_normalizer package
    """

    with open(file_path, "rb") as f:
        data = f.read()
        encoding_for_input = charset_normalizer.detect(data)

    is_confident = encoding_for_input["confidence"] == 1
    if not is_confident:
        logger.warning("Be careful, the detected file encoding for:")
        logger.warning(f"{file_path}")
        logger.warning(r"has less than 100% confidence")

    return encoding_for_input["encoding"]


def read_delim(file_path, cast_d_type="string"):
    """
    reads in a tabular file (ie spreadsheet) after detecting
    encoding and file extension without any type casting.

    currently supports csv and tsv

    defaults to not casting values (ie all columns are string dtypes)
    and not parsing strings into NA values (eg "" is kept as "")

    Returns a pandas dataframe
    """
    ext = Path(file_path).suffix
    if ext == ".csv":
        sep = ","
    elif ext == ".tsv":
        sep = "\t"
    else:
        raise ValueError("Delimited file must be csv or tsv")

    encoding = detect_file_encoding(file_path)
    file_encoding = pd.read_csv(
        file_path, sep=sep, encoding=encoding, dtype=cast_d_type, keep_default_na=False
    )

    return file_encoding


def read_data_from_json_file(input_file: str) -> Dict:
    """Loads the data from a json input file"""
    data = json.loads(Path(input_file).read_text())
    return data


def get_schema(data_or_path, schema_type: str):
    """
    Get the schema for the specified schema_type.
    Use the input_file suffix for "auto".

    Args:
        data_or_path: json dictionary or path to input file
        schema_type (str): the type of the schema that can be validated against.
            Allowed values for now are “csv”, “tsv”, “json” and “auto”. Defaults to “auto”
    Returns:
        schema (dict)
    """

    if isinstance(data_or_path, (str, os.PathLike)):
        dictionary_type = Path(data_or_path).suffix.replace(".", "")
    elif isinstance(data_or_path, dict):
        dictionary_type = "json"
    elif isinstance(data_or_path, list):
        dictionary_type = "csv"
    else:
        logger.error("Cannot get schema. Input is not path or dict or list")
        raise ValueError("Input should be path or dict or list")

    if schema_type == "csv" or (schema_type == "auto" and dictionary_type == "csv"):
        schema = {"type": "array", "items": CSV_SCHEMA}
        return schema
    if schema_type == "tsv" or (schema_type == "auto" and dictionary_type == "tsv"):
        schema = {"type": "array", "items": CSV_SCHEMA}
        return schema
    if schema_type == "json" or (schema_type == "auto" and dictionary_type == "json"):
        schema = JSON_SCHEMA
        return schema

    return None



import re
from os import PathLike
from pathlib import Path

import pandas as pd

from cdislogging import get_logger
from heal.vlmd.config import JSON_SCHEMA
from heal.vlmd.extract import utils
from heal.vlmd.extract.redcap_csv_dict_conversion import convert_redcap_csv
from heal.vlmd.utils import has_redcap_headers
from heal.vlmd.validate.utils import read_delim


class RedcapExtractionError(Exception):
    pass


logger = get_logger("csv-conversion", log_level="info")


def _parse_string_objects(
    tbl_csv: pd.DataFrame, field_properties: dict
) -> pd.DataFrame:
    """Parse string objects and array to create the dict (json) instance"""
    tbl_json = tbl_csv.copy()
    for column_name in tbl_json.columns.tolist():
        # NOTE: the below methodology uses the schema to instruct how to convert to json.
        field_prop_name = utils.find_prop_name(column_name, field_properties)
        field_prop = field_properties.get(field_prop_name)
        if field_prop:
            if field_prop["type"] == "object":
                tbl_json[column_name] = tbl_csv[column_name].apply(
                    utils.parse_dictionary_str, item_sep="|", key_val_sep="="
                )

            elif field_prop["type"] == "array":
                tbl_json[column_name] = tbl_csv[column_name].apply(
                    utils.parse_list_str, item_sep="|"
                )
        else:
            # columns not included in schema ('custom' or 'other')
            # CSV header should have format 'custom.<KEY>'
            group = re.match(r"^custom\.(.+)$", column_name)
            if group:
                key = group[1]
                if "custom" not in tbl_json:
                    tbl_json["custom"] = [{}] * len(tbl_json)
                for i in range(len(tbl_csv)):
                    value = tbl_csv[column_name].iloc[i]
                    if value:
                        tbl_json.at[i, "custom"] = {key: value}
            else:
                # May throw an error unless schema has '"additionalProperties": true'
                tbl_json[column_name] = tbl_csv[column_name]

    return tbl_json


def convert_datadict_csv(
    csv_template: str,
    data_dictionary_props: dict,
    rename_map: dict = None,
    recode_map: dict = None,
    drop_list: dict = None,
    item_sep: str = "|",
    key_val_sep: str = "=",
) -> dict:
    """
    Converts a CSV conforming to HEAL specifications (but see 2 additional notes below)
    into a HEAL-specified data dictionary in both csv format and json format.

    Converts an in-memory data dictionary or a path to a data dictionary file into a HEAL-specified tabular template by:
        1. Adding missing fields, and
        2. Converting fields from a specified mapping.
            NOTE: currently this mapping is only float/num to number or text/char to string (case insensitive)
                In future versions, there will be a specified module for csv input mappings.

    Args

        csv_template : str or path-like or an object that can be inferred as data by frictionless's Resource class.
            Data or path to data with the data being a tabular HEAL-specified data dictionary.
            This input can be any data object or path-like string excepted by a frictionless Resource object.
        data_dictionary_props : dict
            The HEAL-specified data dictionary properties.
        rename_map: A mapping of source (current) column headers to target (desired -- conforming to CVS HEAL spec)
            column headers
        recode_map: A mapping of values for each column in HEAL spec, eg,
            {..."column_name":{"old_value":"new_value"...}...}
        drop_list: a list of variables to drop from headers before processing
        item_sep:str (default:"|") Used to split stringified items (in objects and arrays)
        key_val_sep:str (default:"=") Used to split stringified each key-value pair

    Returns
        A dictionary with two keys:
            - 'template_json': the HEAL-specified JSON object.
            - 'template_csv': the HEAL-specified tabular template.

    """

    # get transforms
    # cast numbers explicitly based on schema
    # this is needed in case there is only one record in a string column that is a number (ie don't want to convert)
    def infer_delim(series: pd.Series, char_list: list, firstmatch: bool):
        """infer the delimiter by the highest frequency by row higest occuring character
        can either take only the first match or all matches for each row"""
        combined_pattern = "|".join(map(re.escape, char_list))

        try:
            # get most frequent delimiter per row
            most_freq_chars = (
                series.rename_axis("row")
                .str.extractall(f"({combined_pattern})")
                .squeeze(
                    "rows"
                )  # only 1 row, dont want to squeeze to scalar so spec rows
                .pipe(
                    lambda s: s.unstack("match")[0] if firstmatch else s
                )  # if first match specified
                .groupby("row")
                .value_counts()
                .unstack()
                .idxmax(axis=1)
            )
        except KeyError:
            return None

        try:
            # infer delim by character with most per row max frequencies
            inferred_delim = most_freq_chars.value_counts().sort_values().index[-1]
        except IndexError:
            return None

        return inferred_delim

    if isinstance(csv_template, (str, PathLike)):
        template_tbl = read_delim(str(Path(csv_template)))
    else:
        template_tbl = pd.DataFrame(csv_template)

    # If REDCap then get dictionary and return or raise RedcapExtractionError.
    column_names = template_tbl.columns
    if has_redcap_headers(column_names):
        logger.debug("File appears to have REDCap headers. Ready to convert.")
        try:
            converted_dict = convert_redcap_csv(template_tbl)
        except Exception as err:
            logger.error("Error in extracting REDCap dictionary")
            logger.error(err)
            raise RedcapExtractionError(str(err))
        return converted_dict
    else:
        logger.debug("File is CSV dictionary, not REDCap dictionary.")

    # Regular CSV, not REDCap.
    if not rename_map:
        rename_map = {}

    if not recode_map:
        recode_map = {}

    if not drop_list:
        drop_list = []

    slugify = lambda s: s.strip().lower().replace("_", "-").replace(" ", "-")
    # flattened properties
    field_properties = utils.flatten_properties(
        JSON_SCHEMA["properties"]["fields"]["items"]["properties"]
    )

    # init to-be formatted tables
    tbl_csv = template_tbl.copy()
    # transform each column with slugified mappings, harmonizing delims (if array, object)
    for column_name in tbl_csv.columns.tolist():
        slugified_col = slugify(column_name)
        new_column_name = column_name

        # rename based on slugified names or original col names
        new_column_name = rename_map.get(slugified_col) or rename_map.get(column_name)
        if new_column_name:
            tbl_csv.rename(columns={column_name: new_column_name}, inplace=True)
        else:
            new_column_name = column_name

        # recode slugified old names to accepted names
        if new_column_name in recode_map.keys():
            tbl_csv[new_column_name] = (
                tbl_csv[new_column_name]
                .apply(slugify)
                .replace(recode_map[new_column_name])
            )

        if new_column_name in drop_list:
            del tbl_csv[new_column_name]

        # NOTE: the below methodology uses the schema to instruct how to convert to json.
        field_prop_name = utils.find_prop_name(new_column_name, field_properties)
        field_prop = field_properties.get(field_prop_name)
        # infer delimiters of stringified lists (note: stringified lists are identified from schema)

        if field_prop_name:
            if field_prop["type"] == "integer":
                tbl_csv[new_column_name] = tbl_csv[new_column_name].apply(
                    lambda s: int(float(s)) if s else s
                )
            elif field_prop["type"] == "number":
                tbl_csv[new_column_name] = tbl_csv[new_column_name].astype(float)
            elif field_prop["type"] == ["integer", "number"]:
                tbl_csv[new_column_name] = tbl_csv[new_column_name].apply(
                    lambda s: float(s) if s else s
                )
            elif field_prop["type"] == "object":
                possible_key_val = ["=", ":"]
                possible_list = [";", "|"]
                key_val_sep = (
                    infer_delim(
                        tbl_csv[new_column_name], possible_key_val, firstmatch=True
                    )
                    or "="
                )
                item_sep = (
                    infer_delim(
                        tbl_csv[new_column_name], possible_list, firstmatch=False
                    )
                    or "|"
                )

                tbl_csv[new_column_name] = (
                    tbl_csv[new_column_name]
                    .str.replace(key_val_sep, "=")
                    .str.replace(item_sep, "|")
                )

            elif field_prop["type"] == "array":
                possible_list = [";", "|"]
                item_sep = (
                    infer_delim(
                        tbl_csv[new_column_name], possible_list, firstmatch=False
                    )
                    or "|"
                )
                tbl_csv[new_column_name] = tbl_csv[new_column_name].replace(
                    item_sep, "|"
                )

    tbl_json = _parse_string_objects(tbl_csv, field_properties)

    # drop all custom columns (as I have nested already)
    tbl_json.drop(columns=tbl_json.filter(regex="^custom\\.").columns, inplace=True)

    # refactor (i.e., cascade, move up to root) properties if present in all records
    refactored_props, tbl_json = utils.refactor_field_props(
        tbl_json, schema=JSON_SCHEMA
    )

    # data dictionary root level properties
    data_dictionary_props_csv = dict(data_dictionary_props)
    data_dictionary_props_json = {
        **data_dictionary_props,
        **utils.unflatten_from_json_path(refactored_props.to_dict()),
    }

    # create the data dictionary objects
    fields_json = [
        utils.unflatten_from_json_path(record)
        for record in tbl_json.to_dict(orient="records")
    ]
    template_json = dict(**data_dictionary_props_json, fields=fields_json)

    fields_csv = tbl_csv.to_dict(orient="records")
    template_csv = dict(**data_dictionary_props_csv, fields=fields_csv)

    return {"template_json": template_json, "template_csv": template_csv}



import re
from os.path import isfile
from pathlib import Path

import jsonschema
from cdislogging import get_logger
from jsonschema import ValidationError

from heal.vlmd import ExtractionError, vlmd_validate
from heal.vlmd.config import (
    ALLOWED_FILE_TYPES,
    ALLOWED_INPUT_TYPES,
    ALLOWED_OUTPUT_TYPES,
    TOP_LEVEL_PROPS,
)
from heal.vlmd.extract.conversion import convert_to_vlmd
from heal.vlmd.extract.csv_dict_conversion import RedcapExtractionError
from heal.vlmd.file_utils import get_output_filepath, write_vlmd_dict
from heal.vlmd.validate.validate import file_type_to_fxn_map
from heal.vlmd.utils import add_types_to_props
from heal.vlmd.validate.utils import get_schema


logger = get_logger("extract", log_level="info")


def set_title_if_missing(file_type: str, title: str, converted_dict: dict) -> dict:
    """
    Dictionary from json input should have an existing title or a user-supplied title.
    Dictionary from non-json input should set the title from
    standardsMappings[0].instrument.title or from the user supplied title.
    """

    existing_title = converted_dict.get("title")
    default_csv_title = TOP_LEVEL_PROPS.get("title")

    if file_type == "json":
        if existing_title is not None and existing_title != default_csv_title:
            logger.info(f"Converted dict already has a title '{existing_title}'")
            return converted_dict

        if title is not None:
            logger.debug(f"JSON dictionary setting user-defined title '{title}'")
            converted_dict["title"] = title
            return converted_dict

    else:
        instrument_title = None
        try:
            standards_mappings = converted_dict.get("standardsMappings")
            if standards_mappings and len(standards_mappings) >= 1:
                instrument_title = standards_mappings[0].get("instrument").get("title")
        except Exception as err:
            logger.warning("standardsMapping does not have 'instrument.title'")

        if title is None and instrument_title is None:
            message = "Title must be supplied when extracting from non-json to json"
            logger.error(message)
            raise ExtractionError(message)

        if instrument_title is not None:
            logger.debug(
                f"Converted CSV dictionary setting title to instrument title '{instrument_title}'"
            )
            converted_dict["title"] = instrument_title
            return converted_dict

        if title is not None and (
            existing_title is None or existing_title == default_csv_title
        ):
            logger.debug(
                f"Converted CSV dictionary setting user-defined title '{title}'"
            )
            converted_dict["title"] = title
            return converted_dict

    return converted_dict


def vlmd_extract(
    input_file: str,
    title: str = None,
    file_type: str = "auto",
    output_dir: str = ".",
    output_type: str = "json",
    include_all_fields: bool = True,
) -> bool:
    """
    Extract a HEAL compliant csv and json format VLMD data dictionary
    from the input file.

    Args:
        input_file (str): the path of the input HEAL VLMD file to be extracted
            into HEAL-compliant VLMD file(s).
        title (str): the root level title of the dictionary (required if extracting
            from csv to json)
        file_type (str): the type of the input file that will be extracted into a
            HEAL-compliant VLMD file.
            Allowed values are "auto", “csv”, "dataset_csv", "dataset_tsv",
                "json", "tsv", and "redcap".
            where "redcap" is a csv file of a REDCap dictionary export.
            Defaults to “auto”.
        output_dir (str): the directory of where the extracted VLMD file will
            be written. Defaults to “.”
        output_type (str): format of dictionary to write: "csv" or "json".
            The default is "json".
        include_all_fields (bool): If true then csv dictionaries extracted from
            csv datasets will include columns for all fields in the schema.
            Useful for generating a template that can be manually updated.
            Default = True.

    Returns:
        True if the input is valid and is successfully converted and written.
        Raises ExtractionError in input VLMD is not valid or could not be converted.
    """

    logger.info(
        f"Extracting VLMD file '{input_file}' with input file_type '{file_type}'"
    )

    file_suffix = Path(input_file).suffix.replace(".", "")
    if file_suffix not in ALLOWED_INPUT_TYPES:
        message = f"Input file must be one of {ALLOWED_INPUT_TYPES}"
        logger.error(message)
        raise ExtractionError(message)

    if not isfile(input_file):
        message = f"Input file does not exist: {input_file}"
        logger.error(message)
        raise ExtractionError(message)

    if file_type not in ALLOWED_FILE_TYPES:
        message = f"File type must be one of {ALLOWED_FILE_TYPES}"
        logger.error(message)
        raise ExtractionError(message)

    type_is_auto = False
    if file_type == "auto":
        # Set as data dictionary
        file_type = file_suffix
        logger.debug(f"Changing file_type from 'auto' to '{file_type}'")
        type_is_auto = True

    if output_type not in ALLOWED_OUTPUT_TYPES:
        message = f"Unrecognized output_type '{output_type}' - should be in {ALLOWED_OUTPUT_TYPES}"
        logger.error(message)
        raise ExtractionError(message)

    if title is not None and re.match(r"^\s*$", title):
        message = f"Empty title is not allowed"
        logger.error(message)
        raise ExtractionError(message)

    # input json file require explicit conversion and post validation steps
    if file_type == "json":
        file_convert_function = file_type_to_fxn_map.get(file_type)
        data_dictionary_props = {}
        try:
            logger.debug("Ready to convert json input to VLMD")
            data_dictionaries = convert_to_vlmd(
                input_filepath=input_file,
                input_type=file_convert_function,
                data_dictionary_props=data_dictionary_props,
            )
            if output_type == "json":
                converted_dictionary = data_dictionaries["template_json"]
                logger.debug(
                    f"Ready to validate converted dict with output type '{output_type}'"
                )
                is_valid = vlmd_validate(
                    converted_dictionary,
                    file_type=file_type,
                    output_type=output_type,
                    return_converted_output=False,
                )
                logger.debug(f"Converted dictionary is valid: {is_valid}")
            else:
                converted_dictionary = data_dictionaries["template_csv"]["fields"]
        except Exception as err:
            logger.error(f"Error in extracting JSON dictionary from {input_file}")
            logger.error(err)
            raise ExtractionError(str(err))

    elif file_type in ["csv", "redcap", "tsv"]:
        try:
            # csv files are converted as part of validate
            converted_dictionary = vlmd_validate(
                input_file,
                file_type=file_type,
                output_type=output_type,
                return_converted_output=True,
            )
        except RedcapExtractionError as err:
            logger.error("Error in extracting REDCap dictionary")
            if type_is_auto:
                logger.info(
                    "File_type was 'auto' but input identified as REDCap. Skipping fallback to dataset."
                )
            raise ExtractionError(str(err))
        except ValidationError as err:
            logger.error(
                f"Error in validating and extracting dictionary from {input_file}"
            )
            logger.error(err.message)
            if type_is_auto:
                file_type = f"dataset_{file_suffix}"
                logger.info(
                    f"Will now attempt to extract file '{input_file}' as '{file_type}'"
                )
            else:
                raise ExtractionError(str(err.message))
        except Exception as err:
            logger.error(f"Error in extracting dictionary from {input_file}")
            logger.error(err)
            if type_is_auto:
                file_type = f"dataset_{file_suffix}"
                logger.info(
                    f"Will attempt to extract file '{input_file}' as '{file_type}'"
                )
            else:
                raise ExtractionError(str(err))

    # If data file then try conversion
    if file_type in ["dataset_csv", "dataset_tsv"]:
        file_convert_function = file_type_to_fxn_map.get(file_type)
        data_dictionary_props = {}
        try:
            logger.debug("Ready to convert csv data set input to VLMD")
            data_dictionaries = convert_to_vlmd(
                input_filepath=input_file,
                input_type=file_convert_function,
                data_dictionary_props=data_dictionary_props,
                include_all_fields=include_all_fields,
            )

            if output_type == "json":
                converted_dictionary = data_dictionaries["template_json"]
            else:
                converted_dictionary = data_dictionaries["template_csv"]["fields"]

        except ValidationError as err:
            logger.error(
                f"Error in validating and extracting dataset from {input_file}"
            )
            logger.error(err.message)
            raise ExtractionError(str(err.message))
        except Exception as err:
            logger.error(f"Error in extracting dictionary from dataset {input_file}")
            logger.error(err)
            raise ExtractionError(str(err))

        schema_type = "csv"
        schema = get_schema(converted_dictionary, schema_type=output_type)
        if output_type == "csv":
            schema = add_types_to_props(schema)
        if schema is None:
            message = f"Could not get schema for type = {schema_type}"
            logger.error(message)
            raise ValueError(message)
        try:
            logger.debug(f"Validating converted dictionary")
            jsonschema.validate(instance=converted_dictionary, schema=schema)
        except jsonschema.ValidationError as err:
            logger.error("Error in validating converted dictionary")
            raise err
        logger.debug("Converted dictionary is valid")

    if output_type == "json":
        converted_dictionary = set_title_if_missing(
            file_type=file_type, title=title, converted_dict=converted_dictionary
        )
        if converted_dictionary.get("title") is None:
            logger.error("JSON dictionary is missing 'title'")
            raise ExtractionError("JSON dictionary is missing 'title'")

    # write to file
    output_filepath = get_output_filepath(
        output_dir, input_file, output_type=output_type
    )
    logger.info(f"Writing converted dictionary to {output_filepath}")
    try:
        write_vlmd_dict(converted_dictionary, output_filepath, file_type=output_type)
    except Exception as err:
        logger.error("Error in writing converted dictionary")
        logger.error(err)
        raise ExtractionError("Error in writing converted dictionary")

    return True



import json
from pathlib import Path

# file prefix
OUTPUT_FILE_PREFIX = "heal-dd"

# file suffixes
ALLOWED_INPUT_TYPES = ["csv", "tsv", "json"]
ALLOWED_FILE_TYPES = [
    "auto",
    "csv",
    "dataset_csv",
    "dataset_tsv",
    "json",
    "redcap",
    "tsv",
]
ALLOWED_SCHEMA_TYPES = ["auto", "csv", "json", "tsv"]
ALLOWED_OUTPUT_TYPES = ["csv", "json"]

# schemas
schema_dir_path = Path(__file__).parents[1].joinpath("vlmd/schemas")

csv_schema_path = schema_dir_path.joinpath("heal_csv.json")
with open(csv_schema_path, "r") as schema_file:
    CSV_SCHEMA = json.load(schema_file)

json_schema_path = schema_dir_path.joinpath("heal_json.json")
with open(json_schema_path, "r") as schema_file:
    JSON_SCHEMA = json.load(schema_file)

# schema
JSON_SCHEMA_VERSION = JSON_SCHEMA.get("version", "0.3.2")
# The title is a default title used in the validation process.
# It will get overwritten by a user-specified title in the extraction process.
TOP_LEVEL_PROPS = {
    "schemaVersion": JSON_SCHEMA_VERSION,
    "title": "HEAL Data Dictionary",
}



"""
This module includes an external file retriever function intended to be called
by the external_files_download module in the Gen3-SDK.

The retriever function sends requests to the Harvard Dataverse for downloading studies or files.
TODO: QDR and Harvard Dataverse use the same Dataverse API, however, we do NOT need to use WTS token to access with Harvard
The Dataverse documentation describes how to download studies
https://guides.dataverse.org/en/latest/api/dataaccess.html#basic-download-by-dataset

"""

from pathlib import Path
from typing import Dict, List
from heal.utils import unpackage_object, get_id, download_from_url

from cdislogging import get_logger
from gen3.tools.download.drs_download import DownloadStatus

logger = get_logger("__name__", log_level="debug")


def get_harvard_dataverse_files(
    wts_hostname: str, auth, file_metadata_list: List, download_path: str = "."
) -> Dict:
    """
    Retrieves external data from the Harvard Dataverse.

    Args:
        wts_hostname (str): hostname for commons with wts (not being used at this moment)
        auth (Gen3Auth): auth for commons with wts (not being used at this moment)
        file_metadata_list (List of Dict): list of studies or files
        download_path (str): path to download files and unpack

    Returns:
        Dict of download status
    """
    if not Path(download_path).exists():
        logger.critical(f"Download path does not exist: {download_path}")
        return None

    completed = {}
    logger.debug(f"Input file metadata list={file_metadata_list}")

    for file_metadata in file_metadata_list:
        id = get_id(file_metadata)
        if id is None:
            logger.warning(
                f"Could not find 'study_id' or 'file_id' in metadata {file_metadata}"
            )
            continue
        logger.info(f"ID = {id}")
        completed[id] = DownloadStatus(filename=id, status="pending")

        download_url = get_download_url_for_harvard_dataverse(file_metadata)
        if download_url is None:
            logger.critical(f"Could not get download_url for {id}")
            completed[id].status = "invalid url"
            continue

        logger.debug(f"Ready to send request to download_url: GET {download_url}")
        downloaded_file = download_from_url(
            api_url=download_url,
            headers=None,
            download_path=download_path,
        )
        if downloaded_file is None:
            completed[id].status = "failed"
            continue

        if downloaded_file.endswith("zip"):
            # unpack if download is zip file
            try:
                logger.debug(f"Ready to unpack {downloaded_file}.")
                unpackage_object(filepath=downloaded_file)
            except Exception as e:
                logger.critical(f"{id} had an issue while being unpackaged: {e}")
                completed[id].status = "failed"

            completed[id].status = "downloaded"
            # remove the zip file
            Path(downloaded_file).unlink()
        else:
            completed[id].status = "downloaded"

    if not completed:
        return None
    return completed


def get_download_url_for_harvard_dataverse(file_metadata: Dict) -> str:
    """
    Get the download url for Harvard Dataverse.

    Args:
        file_metadata (Dict)

    Returns:
        url, None if there are errors
    """
    base_url = "https://dataverse.harvard.edu/api/access"
    if "use_harvard_staging" in file_metadata and bool(
        file_metadata["use_harvard_staging"]
    ):
        base_url = "https://demo.dataverse.org/api/access"
    if "study_id" in file_metadata:
        url = f"{base_url}/dataset/:persistentId/?persistentId={file_metadata.get('study_id')}"
    elif "file_id" in file_metadata:
        url = f"{base_url}/datafile/{file_metadata.get('file_id')}"
    else:
        url = None

    return url


def is_valid_harvard_file_metadata(file_metadata: Dict) -> bool:
    """
    Check that the file_metadata has the required keys:
    'study_id' or 'file_id'.

    Args:
        file_metadata (Dict)

    Returns:
        True if valid file_metadata object.
    """
    if not isinstance(file_metadata, dict):
        logger.critical(f"Invalid metadata - item is not a dict: {file_metadata}")
        return False
    if "study_id" not in file_metadata and "file_id" not in file_metadata:
        logger.critical(
            f"Invalid metadata - missing required Harvard Dataverse keys {file_metadata}"
        )
        return False
    return True



from functools import partial

import pandas as pd
import pandas.api.types as pdt
import visions as v

# for declarative API example used as reference see:
# https://github.com/dylan-profiler/visions/blob/develop/examples/declarative_typeset.py


# for Categorical vision type reference, see ydata-profiling:
# https://github.com/ydataai/ydata-profiling/blob/develop/src/ydata_profiling/model/typeset.py
# state param needed for visions graph traversal
class contains:
    def is_category(series, state):
        return isinstance(series.dtype, pd.CategoricalDtype)
        # return pdt.is_categorical_dtype(series)

    def is_boolean(series, state):
        return pdt.is_bool_dtype(series)


class _transformers:
    def to_category(series, state):
        return pd.Categorical(series)


class _relationships:
    def type_is_category(series, k, threshold):
        nunique = series.nunique()
        few_enough_cats = nunique <= k
        low_enough_thresh = (nunique / series.size) < threshold
        is_not_boolean = not contains.is_boolean(series, None)
        return few_enough_cats and low_enough_thresh and is_not_boolean


class inference_relations:
    def type_to_category(
        related_types=[v.Integer, v.Float, v.String], k=5, threshold=0.2
    ):
        relationships = []
        for vision_type in related_types:
            relation = dict(
                related_type=vision_type,
                relationship=lambda series, state: partial(
                    _relationships.type_is_category, k=k, threshold=threshold
                )(series),
                transformer=_transformers.to_category,
            )
            relationships.append(relation)
        return relationships


# types in the CompleteSet but not in StandardSet
## just manually add additions so additional dependencies
## for unused types not necessary

# {Count,
#  Date,
#  EmailAddress,
#  File,
#  Geometry,
#  IPAddress,
#  Image,
#  Ordinal,
#  Path,
#  Time,
#  URL,
#  UUID}

# TODO: make function to allow specification of params (eg k and threshold). Currently using sensiable default (see typesets_relations.py)
# TODO: should we call this something more "frictionless-y" like "Enum"?
Categorical = v.create_type(
    "Categorical",
    identity=[v.Generic],
    contains=contains.is_category,
    inference=inference_relations.type_to_category(),
)

typeset_original = v.StandardSet() - v.Categorical + v.Date + v.Time
typeset_with_categorical = typeset_original + Categorical

typeset_mapping = {
    "Integer": "integer",
    "Float": "number",
    "String": "string",
    "Time": "time",
    "TimeDelta": "duration",
    "DateTime": "datetime",
    "Boolean": "boolean",
}


def infer_frictionless_fields(
    df,
    typesets=[typeset_original, typeset_with_categorical],
    typeset_mapping=typeset_mapping,
):
    """
    Takes in a dataframe
    and infers types by iterating through typesets.


    NOTE: This multiple typeset iteration was a solution to
    correctly inferring types within categoricals
    (eg, integer categoricals from a float column.
    best solution may be to look into ordering inferences but
    this does the same job.)
    """

    # TODO: infer formats with extended dtypes (eg url, email etc)
    for typeset in typesets:
        df, typepaths, _ = typeset.infer(
            df
        )  # typepaths is list of the visions graph traversal - last item is casted type
    fields = []

    for col, typepath in typepaths.items():
        field = {"name": col}
        if len(typepath) == 1:
            continue
        type_final = str(typepath[-1])
        if type_final == "Categorical":
            # TODO: see visions PR https://github.com/dylan-profiler/visions/issues/160 -- best way
            # would probably be to use networkx graph. seems like this may be a good feature to add
            # to visions package
            type_second_to_final = str(typepath[-2])
            field["type"] = typeset_mapping.get(type_second_to_final, "any")
            # enums for inferred categoricals
            field["constraints"] = {"enum": list(df[col].cat.categories)}
        else:
            field["type"] = typeset_mapping.get(str(type_final), "any")

        if field["type"] == "any":
            print(field["name"])
        fields.append(field)

    return fields



import json
from os import PathLike
from pathlib import Path
from typing import Dict

import jsonschema
from cdislogging import get_logger

from heal.vlmd.config import ALLOWED_SCHEMA_TYPES
from heal.vlmd.validate.utils import get_schema

logger = get_logger("json-validator", log_level="info")


def read_data_from_json_file(input_file: str) -> Dict:
    """Loads the data from a json input file"""
    data = json.loads(Path(input_file).read_text())
    return data


def vlmd_validate_json(data_or_path, schema_type: str) -> bool:
    """
    Validate json file against specified schema type.

    Args:
        data_or_path: the path of the input HEAL VLMD file that to be validated
        schema_type (str): the type of the schema that can be validated against.
            Allowed values for now are “csv”, “tsv”, “json” and “auto”. Defaults to “auto”

    Returns:
        True for valid data and schema.
        Raises jsonschema.ValidationError if data is not valid
        or jsonschema.SchemaError if schema is not valid
    """

    if schema_type not in ALLOWED_SCHEMA_TYPES:
        raise ValueError(f"Schema type must be in {ALLOWED_SCHEMA_TYPES}")

    if isinstance(data_or_path, (str, PathLike)):
        logger.debug("Getting json data from file")
        data = read_data_from_json_file(data_or_path)
    else:
        logger.debug("Getting json data from object")
        data = data_or_path

    schema = get_schema(data_or_path, schema_type)
    if schema is None:
        raise ValueError(f"Could not get schema for type = {schema_type}")
    logger.debug("Checking schema")
    jsonschema.validators.Draft7Validator.check_schema(schema)

    logger.debug("Validating data")
    jsonschema.validate(instance=data, schema=schema)

    return True



"""
CSV data to HEAL VLMD conversion
"""

from heal.vlmd.extract.json_dict_conversion import convert_template_json
from heal.vlmd.mappings import typesets
from heal.vlmd.validate.utils import read_delim


def convert_dataset_csv(file_path, data_dictionary_props={}):
    """
    Takes a CSV file containing data (not metadata) and
    infers each of it's variables data types and names.
    These inferred properties are then outputted as partially-completed
    HEAL variable level metadata files. That is, it outputs the `name` and `type` property.

    NOTE: this will be an invalid file as `description` is required
    for each variable. However, this serves as a great way to start
    the basis of a VLMD submission.
    """
    df = read_delim(file_path)
    data_dictionary = data_dictionary_props.copy()
    fields = typesets.infer_frictionless_fields(df)
    data_dictionary["fields"] = fields

    package = convert_template_json(data_dictionary)
    return package



import collections
import json
from os import PathLike
from pathlib import Path

import pandas as pd

from cdislogging import get_logger
from heal.vlmd.config import JSON_SCHEMA
from heal.vlmd.extract import utils

logger = get_logger("json-conversion", log_level="info")


def convert_template_json(
    json_template,
    data_dictionary_props: dict = None,
    fields_name: str = "fields",
) -> dict:
    """
    Converts a JSON file or dictionary conforming to HEAL specifications
    into a HEAL-specified data dictionary in both csv format and json format.

    Converts in-memory data or a path to a data dictionary file.

    If data_dictionary_props is specified, any properties passed in will be
    overwritten.

    Args
        json_template : str or path-like or an object that can be inferred as data by frictionless's Resource class.
            Data or path to data with the data being a tabular HEAL-specified data dictionary.
            This input can be any data object or path-like string excepted by a frictionless Resource object.
        data_dictionary_props : dict
            The HEAL-specified data dictionary properties.

    Returns
        A dictionary with two keys:
            - 'template_json': the HEAL-specified JSON object.
            - 'template_csv': the HEAL-specified tabular template.

    """

    if isinstance(json_template, (str, PathLike)):
        logger.debug(f"Getting data from path to JSON file '{json_template}'")
        json_template_dict = json.loads(Path(json_template).read_text())
    elif isinstance(json_template, collections.abc.MutableMapping):
        json_template_dict = json_template
    else:
        raise ValueError(
            "json_template needs to be either dictionary-like or a path to a json"
        )

    existing_title = json_template_dict.get("title")

    if data_dictionary_props:
        for prop_name, prop in data_dictionary_props.items():
            # determine if you should write or overwrite the root level data dictionary props
            if not json_template_dict.get(prop_name):
                write_prop = True
            elif prop and prop != json_template_dict.get(prop_name):
                write_prop = True
            else:
                write_prop = False

            if write_prop:
                json_template_dict[prop_name] = prop

    fields_json = json_template_dict.pop(fields_name)
    data_dictionary_props = json_template_dict

    fields_schema = JSON_SCHEMA["properties"]["fields"]["items"]
    flattened_fields = pd.DataFrame(
        [utils.flatten_to_json_path(f, fields_schema) for f in fields_json]
    )
    flattened_data_dictionary_props = pd.Series(
        utils.flatten_to_json_path(data_dictionary_props, JSON_SCHEMA)
    )
    flattened_and_embedded = utils.embed_data_dictionary_props(
        flattened_fields, flattened_data_dictionary_props, JSON_SCHEMA
    )
    tbl_csv = (
        flattened_and_embedded.fillna("")
        .map(
            lambda v: utils.join_dict_items(v)
            if isinstance(v, collections.abc.MutableMapping)
            else v
        )
        .map(
            lambda v: utils.join_iter(v)
            if isinstance(v, collections.abc.MutableSequence)
            else v
        )
    )
    fields_csv = tbl_csv.to_dict(orient="records")

    template_json = {**data_dictionary_props, "fields": fields_json}
    if existing_title:
        template_json["title"] = existing_title
    template_csv = {**data_dictionary_props, "fields": fields_csv}

    return {"template_json": template_json, "template_csv": template_csv}



"""

This module contains all property name and value mappings
for both the json and csv template. Note, the json property names
are represented as json paths. In v0.2 schema versions,csv field property
names must now conform to json path but in prior versions this was not the case.

"""

from .names import rename_map
from .values import recode_map



import logging

import cdislogging
import click

import heal.cli.vlmd as vlmd


@click.group()
@click.option(
    "--silent",
    "silent",
    is_flag=True,
    default=False,
    help="don't show ANY logs",
)
@click.pass_context
def main(ctx, silent):
    """HEAL-Platform SDK Command Line Interface"""
    ctx.ensure_object(dict)

    if silent:
        # we still need to define the logger, the log_level here doesn't
        # really matter b/c we immediately disable all logging
        logger = cdislogging.get_logger("heal_cli", log_level="debug")
        # disables all logging
        logging.disable(logging.CRITICAL)


main.add_command(vlmd.vlmd)
if __name__ == "__main__":
    main()



"""
Functions for each REDCap field type that uses REDCap field info to determine
various pieces of field metadata.

Input assumes a dictionary with {"REDCap field_name": "REDCap value"}

REDCap dictionary fields:

NOTES - large text box for lots of text
DROPDOWN - dropdown menu with options
RADIO - radio buttons with options
CHECKBOX - checkboxes to allow selection of more than one option
FILE - upload a document
CALC - perform real-time calculations
SQL - select query statement to populate dropdown choices
DESCRIPTIVE - text displayed with no data entry and optional image/file attachment
SLIDER - visual analogue scale; coded as 0-100
YESNO - radio buttons with yes and no options; coded as 1, Yes | 0, No
TRUEFALSE - radio buttons with true and false options; coded as 1, True | 0, False
"""

import re

from cdislogging import get_logger

from heal.vlmd.extract import utils
from heal.vlmd.mappings.redcap_csv_headers import (
    CALC_FIELD_NAME,
    CHOICES_FIELD_NAME,
    CHOICES_LABEL_INPUT,
    SLIDER_FIELD_NAME,
    TEXT_VALID_FIELD_NAME,
)

logger = get_logger("redcap-mapping", log_level="warning")


def _parse_field_properties_from_encodings(encodings_string: str) -> dict:
    """
    Many of the field types have the same logic
    for conversion to data types and just differ
    in presentation (dropbox, radio box) so making
    fxn to support this

    Currently supports strings, ints, and nums for types
    """
    # parse encodings
    field_encodings = utils.parse_dictionary_str(
        encodings_string, item_sep="|", key_val_sep=","
    )
    # get enums
    field_enums = list(field_encodings.keys())
    # interpret type from enums
    for value in field_enums:
        value = value.strip()
        if value.isnumeric():
            try:
                int(value)
                field_type = "integer"
            except ValueError:
                field_type = "number"
        else:
            field_type = "string"

    return {
        "type": field_type,
        "enumLabels": {
            key.strip(): value.strip() for key, value in field_encodings.items()
        },
        "constraints": {"enum": [value.strip() for value in field_enums]},
    }


def map_text(field: dict) -> dict:
    """
    TEXT - single-line text box (for text and numbers)

    Looks at the text validation field, defined by
    'TEXT_VALID_FIELD_NAME' in 'redcap_csv_headers.py'
    """
    integer_pattern = r"^-?\s*\d+\s*$"
    number_pattern = r"^-?\s*\d+(\.\d+)?([eE][-+]?\d+)?\s*$"
    if field.get(TEXT_VALID_FIELD_NAME):
        text_validation = field[TEXT_VALID_FIELD_NAME].lower()
    else:
        text_validation = ""
    field_format = None
    field_pattern = None
    constraints_min = None
    constraints_max = None
    if "datetime" in text_validation:
        field_type = "datetime"
        field_format = "any"
    elif "date" in text_validation:
        field_type = "date"
        field_format = "any"
    elif text_validation == "email":
        field_type = "string"
        field_format = "email"
    elif text_validation == "integer":
        field_type = "integer"
        min_value = field.get("text_valid_min")
        if min_value:
            if re.match(integer_pattern, min_value):
                constraints_min = int(min_value)
            else:
                warning_message = (
                    f"Skipping conversion of min value '{min_value}' for field '"
                    f"{field.get('name')}"
                    "'"
                )
                logger.warning(warning_message)
        max_value = field.get("text_valid_max")
        if max_value:
            if re.match(integer_pattern, max_value):
                constraints_max = int(max_value)
            else:
                warning_message = (
                    f"Skipping conversion of max value '{max_value}' for field '"
                    f"{field.get('name')}"
                    "'"
                )
                logger.warning(warning_message)
    elif text_validation == "alpha_only":
        field_type = "string"
        field_pattern = "^[a-zA-Z]+$"
    elif "number" in text_validation:
        field_type = "number"
        min_value = field.get("text_valid_min")
        if min_value:
            if "comma_decimal" in text_validation:
                min_value.replace(",", ".")
            if re.match(number_pattern, min_value):
                constraints_min = float(min_value)
            else:
                warning_message = (
                    f"Skipping conversion of min value '{min_value}' for field '"
                    f"{field.get('name')}"
                    "'"
                )
                logger.warning(warning_message)
        max_value = field.get("text_valid_max")
        if max_value:
            if "comma_decimal" in text_validation:
                max_value = max_value.replace(",", ".")
            if re.match(number_pattern, max_value):
                constraints_max = float(max_value)
            else:
                warning_message = (
                    f"Skipping conversion of max value '{max_value}' for field '"
                    f"{field.get('name')}"
                    "'"
                )
                logger.warning(warning_message)
    elif text_validation == "phone":
        field_type = "string"
        field_pattern = "^[0-9]{3}-[0-9]{3}-[0-9]{4}$"
    elif text_validation == "postalcode_australia":
        field_type = "string"
        field_pattern = "^[0-9]{4}$"
    elif text_validation == "postalcode_canada":
        field_type = "string"
        field_pattern = "^[A-Z][0-9][A-Z] [0-9][A-Z][0-9]$"
    elif text_validation == "ssn":
        field_type = "string"
        field_pattern = "^[0-9]{3}-[0-9]{2}-[0-9]{4}$"
    elif "time" in text_validation:
        field_type = "time"
        field_format = "any"
    elif text_validation == "vmrn":
        field_type = "string"
        field_pattern = "^[0-9]{10}$"
    elif text_validation == "zipcode":
        field_type = "string"
        field_pattern = "^[0-9]{5}$"
    else:
        field_type = "string"

    props = dict(
        zip(
            ["type", "format", "constraints"],
            [
                field_type,
                field_format,
                {
                    "pattern": field_pattern,
                    "minimum": constraints_min,
                    "maximum": constraints_max,
                },
            ],
        )
    )

    # delete props without values (ie None)
    for prop_name in ["type", "format"]:
        if props[prop_name] is None:
            del props[prop_name]
    # constraints will be cleaned later

    return props


def map_notes(field: dict) -> dict:
    """
    NOTES - large text box for lots of text
    """
    return {"type": "string"}


def map_dropdown(field: dict) -> dict:
    """
    DROPDOWN - dropdown menu with options

    Determined by "options" (ie Choices, Calculations, OR Slider Labels)
    """
    encodings_string = field[CHOICES_FIELD_NAME]
    if not encodings_string:
        error_message = (
            "Missing value in dropdown field '"
            f"{field.get('name')}"
            f"' in column '{CHOICES_LABEL_INPUT}'."
        )
        raise ValueError(error_message)
    return _parse_field_properties_from_encodings(encodings_string)


def map_radio(field: dict) -> dict:
    """
    RADIO - radio buttons with options

    Determined by "options" (ie Choices, Calculations, OR Slider Labels)
    """
    encodings_string = field[CHOICES_FIELD_NAME]
    if not encodings_string:
        error_message = (
            "Missing value in radio field '"
            f"{field.get('name')}"
            f"' in column '{CHOICES_LABEL_INPUT}'."
        )
        raise ValueError(error_message)
    return _parse_field_properties_from_encodings(encodings_string)


def map_checkbox(field: dict) -> dict:
    """
    CHECKBOX	- checkboxes to allow selection of more than one option

    ## Are data from checkbox (choose all that apply) field types handled
    differently from other field types when imported or exported?
    Yes. When your data are exported, each option from a checkbox field
    becomes a separate variable coded 1 or 0 to reflect whether it is checked
    or unchecked. By default, each option is pre-coded 0, so even if you have
    not yet collected any data, you will see 0's for each checkbox option.
    The variable names will be the name of the field followed by the option number.
    So, for example, if you have a field coded as follows:

    Race

    1, Caucasian
    2, African American
    3, Asian
    4, Other

    In your exported dataset, you will have four variables representing the field
    Race that will be set as 0 by default, coded 1 if the option was checked for
    a record. The variable names will consist of the field name, three underscores,
    and the choice value:

    race___1
    race___2
    race___3
    race___4

    Notes:

    when you import data into a checkbox field, you must code it based on the same model
    negative values can be used as the raw coded values for checkbox fields. Due to certain
    limitations, negative values will not work when importing values using the
    Data Import Tool, API and cause problems when exporting data into a statistical
    analysis package. The workaround is that negative signs are replaced by an underscore
    in the export/import-specific version of the variable name (e.g., for a checkbox
    named "race", its choices "2" and "-2" would export as the fields

    race___2
    race____2

    A checkbox field can be thought of as a series of yes/no questions in one field.
    Therefore, a yes (check) is coded as 1 and a no (uncheck) is coded a 0.
    An unchecked response on a checkbox field is still regarded as an answer and
    is not considered missing.
    """
    checkbox_name = field["name"]
    choices = utils.parse_dictionary_str(
        field[CHOICES_FIELD_NAME], item_sep="|", key_val_sep=","
    )
    field_type = "boolean"
    field_enums = ["0", "1"]
    field_encodings = {"0": "Unchecked", "1": "Checked"}

    fields_new = [
        {
            "description": f"[choice={choice}]",
            "title": checkbox_name.title() + ": " + choice,
            "name": checkbox_name
            + "___"
            + re.sub(
                "^\-", "_", val
            ).strip(),  # NOTE: REDCAP changes negative sign to underscore
            "type": field_type,
            "constraints": {"enum": field_enums},
            "enumLabels": field_encodings,
        }
        for val, choice in choices.items()
    ]
    return fields_new


def map_file(field: dict) -> dict:
    return {"type": "string"}


def map_calc(field: dict) -> dict:
    return {"description": f"[calculation: {field[CALC_FIELD_NAME]}]", "type": "number"}


def map_sql(field: dict):
    return None


def map_yes_no(field: dict) -> dict:
    return {
        "type": "boolean",
        "constraints": {"enum": ["0", "1"]},
        "enumLabels": {"0": "No", "1": "Yes"},
    }


def map_true_false(field: dict) -> dict:
    return {
        "type": "boolean",
        "constraints": {"enum": ["0", "1"]},
        "enumLabels": {"0": "False", "1": "True"},
    }


def map_slider(field: dict) -> dict:
    vallist = ["0", "50", "100"]
    lbllist = utils.parse_list_str(field[SLIDER_FIELD_NAME], "|")
    field_encodings = {vallist[i]: lbl for i, lbl in enumerate(lbllist)}
    return {
        "type": "integer",
        "constraints": {"minimum": 0, "maximum": 100},
        "enumLabels": field_encodings,
    }


def map_descriptive(field: dict):
    return None


# not mapping descriptives or sql (TODO: mapping sql?)
type_mappings = {
    "text": map_text,
    "notes": map_notes,
    "dropdown": map_dropdown,
    "radio": map_radio,
    "checkbox": map_checkbox,
    "slider": map_slider,
    "yesno": map_yes_no,
    "truefalse": map_true_false,
    "calc": map_calc,
    "file": map_file,
}



import click
from cdislogging import get_logger

from heal.vlmd.extract.extract import vlmd_extract

logging = get_logger("__name__")


@click.command()
@click.option(
    "--input_file",
    "input_file",
    help="name of file to extract HEAL-compliant VLMD file",
    required=True,
    type=click.Path(writable=True),
)
@click.option(
    "--file_type",
    "file_type",
    help="Type of input file: auto, csv, json, tsv, dataset_csv, dataset_tsv, redcap",
    default="auto",
    type=str,
    show_default=True,
)
@click.option(
    "--title",
    "title",
    help="Root level title for the dictionary (required if extracting from csv to json)",
    default=None,
    type=str,
)
@click.option(
    "--output_dir",
    "output_dir",
    help="directory to write converted dictionary'",
    default=".",
    type=click.Path(writable=True),
    show_default=True,
)
def extract(input_file, title, file_type, output_dir):
    """Extract HEAL-compliant VLMD file from input file"""

    logging.info(f"Extracting VLMD from {input_file}")

    try:
        vlmd_extract(
            input_file,
            title=title,
            file_type=file_type,
            output_dir=output_dir,
        )
    except Exception as e:
        logging.error(f"Extraction error {str(e)}")



import os
import requests
from typing import Dict
from urllib.parse import unquote
import zipfile
from cdislogging import get_logger
from gen3.auth import Gen3Auth
from gen3.tools.download.drs_download import wts_get_token

logger = get_logger("__name__", log_level="debug")


def get_filename_from_headers(headers: Dict) -> str:
    """
    Parse and decode downloaded file name from response headers

    Args:
        headers (dict): response headers

    Returns:
        file name as string
    """
    try:
        file_name = None
        content_response = headers.get("Content-Disposition").split(";")
        for part in content_response:
            # look for UTF-8 encoded file name
            if part.strip().startswith("filename*="):
                file_name = part.split("=", 1)[1].strip()
                if file_name.lower().startswith("utf-8''"):
                    file_name = file_name[7:]
                    file_name = unquote(file_name)
                    break
            elif part.strip().startswith("filename="):
                file_name = part.split("=", 1)[1].strip().strip('"')
                break
        if file_name is None:
            logger.info("Could not parse file name from headers")

    except Exception as e:
        logger.warning("Could not get file name from headers")

    return file_name


def get_idp_access_token(wts_hostname: str, auth: Gen3Auth, file_metadata: Dict) -> str:
    """Get an access token for QDR using a Gen3 commons WTS"""
    try:
        logger.debug("Ready to get auth token")
        wts_access_token = auth.get_access_token()
        logger.debug("Ready to get idp token")
        idp = file_metadata.get("external_oidc_idp")
        idp_access_token = wts_get_token(
            hostname=wts_hostname, idp=idp, access_token=wts_access_token
        )
    except Exception as e:
        logger.critical(f"Could not get token: {e}")
        return None

    return idp_access_token


def unpackage_object(filepath: str):
    """Unpackage the downloaded zip file"""
    with zipfile.ZipFile(filepath, "r") as package:
        package.extractall(os.path.dirname(filepath))


def get_id(file_metadata: Dict) -> str:
    """
    Parse out the object id from the metadata.

    Args:
        file_metadata (Dict)

    Returns:
        string
    """
    id_types = ["study_id", "file_id"]
    for id_type in id_types:
        if id_type in file_metadata:
            return file_metadata.get(id_type)

    return None


def download_from_url(
    api_url: str,
    headers=None,
    download_path: str = ".",
) -> str:
    """
    Retrieve data file (study_id or file_id) from url.
    Save the file based on the filename in the Content-Disposition response header.

    Args:
        api_url (str): url for QDR or Harvard API
        headers (Dict): request headers
        download_path (str): path for saving downloaded zip file

    Returns:
        path to downloaded and renamed file.
    """
    try:
        response = requests.get(url=api_url, headers=headers, stream=True)
        response.raise_for_status()
    except requests.exceptions.Timeout:
        logger.critical(
            f"Was unable to get the download url: {api_url}. Timeout Error."
        )
        return None
    except requests.exceptions.HTTPError as exc:
        logger.critical(f"HTTPError in download {exc}")
        return None
    except requests.exceptions.ConnectionError as exc:
        logger.critical(f"ConnectionError in download {exc}")
        return None
    except Exception as exc:
        logger.critical(f"Error in download {exc}")
        return None
    logger.debug(f"Status code={response.status_code}")
    downloaded_file_name = get_filename_from_headers(response.headers)
    if downloaded_file_name is None:
        downloaded_file_name = api_url.split("/")[-1]
        logger.info(f"Using file name from id in url {downloaded_file_name}")

    if downloaded_file_name.endswith(
        "zip"
    ) and not "application/zip" in response.headers.get("Content-Type"):
        logger.critical("Response headers do not show zipfile content-type")

    total_downloaded = 0
    block_size = 8092  # 8K blocks might want to tune this.
    download_filename = f"{download_path}/{downloaded_file_name}"
    try:
        logger.info(f"Saving download as {download_filename}")
        with open(download_filename, "wb") as file:
            for data in response.iter_content(block_size):
                total_downloaded += len(data)
                file.write(data)
    except IOError as ex:
        logger.critical(f"IOError opening {download_filename} for writing: {ex}")
        return None

    if total_downloaded == 0:
        logger.critical("content-length is 0 and it should not be")
        return None
    logger.debug(f"Download size = {total_downloaded}")

    return download_filename



"""
defines potential source (input) values for the specified field property
that should map directly onto the target (the heal specification) without
any ambiguity
"""


def _flatten(map):
    return {sourcename: item["target"] for item in map for sourcename in item["source"]}


type_map = [
    {"target": "integer", "source": ["int"]},
    {
        "target": "string",
        "source": [
            "str",
            "character",
            "char",
            "text",
            "varchar",
            "alphanumeric",
            "alphanum",
        ],
    },
    {"target": "number", "source": ["num", "float", "decimal", "numeric"]},
    {"target": "boolean", "source": ["bool"]},
]
required_map = [
    {"target": True, "source": ["true", "1", "yes", "y", "required"]},
    {"target": False, "source": ["false", "0", "no", "not required", "n"]},
]
recode_map = {
    "type": _flatten(type_map),
    "constraints.required": _flatten(required_map),
}
